/*
 * Copyright (c) 1997-2006 Erez Zadok,
 * Copyright (c) 2004-2006 Nikolai Joukov,
 * Copyright (c) 2005 Arun M. Krishnakumar,
 * Copyright (c) 2006 Chaitanya Patti,
 * Copyright (c) 2005 Sunil Satnur,
 * Copyright (c) 2006 Avishay Traeger,
 * Copyright (c) 2001-2006 Stony Brook University,
 * Copyright (c) 1997-2000 Columbia University
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: raifctl.c,v 1.20 2006/12/13 17:00:21 kolya Exp $
 */


#define _XOPEN_SOURCE 500

#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include "fist.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <features.h>
#include <ftw.h>

#define RAIF_IOCTL_RULE_ERR -1

static void usage(char *prog) {
	printf("%s add|delete|list|migrate -d directory \"rule\"\n", prog);
	printf("    Rule syntax: \"Type:Position:Regexp:SUnit:Level:BList\"\n");
	printf("    Short rule syntax for deletion: \"Type:Position\"\n");
	printf("\tType    : 'G' for global, 'L' for local\n");
	printf("\tPosition: Position to insert the rule at (rule 0 is tried first)\n");
	printf("\tRegexp  : Regular expression to match filename (e.g., *.c)\n");
	printf("\tSUnit   : Striping unit (multiple of 4KB)\n");
	printf("\tLevel   : RAIF levels (0, 1, 4, 5, or 6)\n");
	printf("\tBList   : Subset of branches to use for this rule (e.g., 0,2,3)\n");
}

int raif_do_migrate(const char *file, const struct stat *sb, int flag, struct FTW *ftw)
{
	int fd, err = 0;

	if (!S_ISDIR(sb->st_mode)) {
		goto out;
	}

	err = chdir(file);
	if (err) {
		perror("RAIF migrate");
		goto out;
	}
	fd = open(file, O_RDONLY);
	if (fd < 0) {
		printf("Error: Unable to open %s\n", file);
		return -1;
	}
	err = ioctl(fd, RAIF_IOCTL_MIGRATE, NULL);

	close(fd);
out:
	return err;
}

void print_warning()
{
	printf("warning: addition/deletion of rules to/from non-empty and\n");
	printf("\tnon-mount-point directories is not yet completely supported\n");
	printf("\tand may result in data corruption\n");
	printf("\t(This warning is printed even if you operate on an empty\n");
	printf("\tmount-point, which is safe.)\n");
}

int main(int argc, char *argv[])
{
	int raifctl_op = RAIF_IOCTL_RULE_ERR;
	int raifmnt_fd, err, global, rulenum;
	char *buf;
	struct stat statbuf;
	struct raif_rulebuf rulebuf;

	if ((argc != 4) && (argc != 5) && (argc != 6)) {
		usage(argv[0]);
		return -1;
	}

	if (!strcmp(argv[1], "add")) {
		raifctl_op = RAIF_IOCTL_RULE_ADD;
		if (argc != 5) {
			usage(argv[0]);
			return -1;
		}
	} else if (!strcmp(argv[1], "delete")) {
		raifctl_op = RAIF_IOCTL_RULE_DEL;
                if (argc != 5) {
                        usage(argv[0]);
                        return -1;
                }
	} else if (!strcmp(argv[1], "list")) {
		raifctl_op = RAIF_IOCTL_RULE_LIST;
                if (argc != 4) {
                        usage(argv[0]);
                        return -1;
                }
	} else if (!strcmp(argv[1], "migrate")) {
		raifctl_op = RAIF_IOCTL_MIGRATE;
                if (argc != 4) {
                        usage(argv[0]);
                        return -1;
                }
	}

	if (raifctl_op == RAIF_IOCTL_RULE_ERR) {
		usage(argv[0]);
		return -1;
	}

	if (strcmp(argv[2], "-d")) {
		usage(argv[0]);
		return -1;
	}

	err = stat(argv[3], &statbuf);
	if (err < 0) {
		printf("Couldnt stat %s\n", argv[3]);
		return -1;
	}
	if (!S_ISDIR(statbuf.st_mode)) {
		printf("%s is not a directory\n", argv[3]);
		return -1;
	}

	raifmnt_fd = open(argv[3], O_RDONLY);
	if (raifmnt_fd < 0) {
		printf("Cannot open directory %s\n", argv[3]);
		return -1;
	}

	if (raifctl_op == RAIF_IOCTL_RULE_ADD) {
		print_warning();
		err = ioctl(raifmnt_fd, raifctl_op, argv[4]);
	} else if (raifctl_op == RAIF_IOCTL_RULE_DEL) {
		print_warning();
		if (argv[4][0] == 'l' || argv[4][0] == 'L') 
			global = 0;
		else if (argv[4][0] == 'g' || argv[4][0] == 'G')
			global= 1;
		else {
			printf("Invalid argument %s\n", argv[4]);
			usage(argv[0]);
			return -1;
		}
		rulenum = atoi(&argv[4][2]);
		if (rulenum < 0){
			printf("Invalid rule number %d\n", rulenum);
			return -1;
		}
		rulebuf.ruletype = global;
		rulebuf.rulenum = rulenum;
		err = ioctl(raifmnt_fd, raifctl_op, &rulebuf);
	} else if (raifctl_op == RAIF_IOCTL_RULE_LIST) { 
		/* List the rules */
		rulebuf.ruletype = 0; /* start with local rules */
		rulebuf.rulenum = 0; /* start with first rule */
		rulebuf.done = 0;
		do {
			err = ioctl(raifmnt_fd, raifctl_op, &rulebuf);
			if (err) {
				perror("\nRAIF rule read");
				goto out;
			}
			printf("%s", rulebuf.buf);
		} while (!rulebuf.done);
		printf("\n");
	} else {
		/* Migrate data */
		return nftw(argv[3], raif_do_migrate, 64, FTW_DEPTH | FTW_CHDIR);
	}
	if (err)
		perror("RAIF ioctl");

out:
	close(raifmnt_fd);
	if (err)
		return -1;
}
